/*
 * keygen.c
 *  WEP Key Generators
 *
 * This program generates WEP keys using de facto standard key
 * generators for 40 and 128 bit keys.
 *
 * I place this code in the public domain.
 * May 2001, Tim Newsham
 * This version was last modified on 110731
 */

/* jmc: slight tweaks - uses md5.c instead of libcrypto */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <openssl/md5.h>

#define PROGNAME            "wepkeygen"
#define REVISION            "110731"
#define WEPKEYSIZE          5
#define WEPSTRONGKEYSIZE    13
#define WEPKEYS             4
#define WEPKEYSTORE         (WEPKEYSIZE * WEPKEYS)

static int strong = 0;

/*
 * generate four subkeys from a seed using the defacto standard
 */
void
wep_seedkeygen(int val, u_char *keys)
{
    int i;

    for(i = 0; i < WEPKEYSTORE; i++) {
        val *= 0x343fd;
        val += 0x269ec3;
        keys[i] = val >> 16;
    }
    return;
}

/*
 * generate one key from a string using the de facto standard
 *
 * resultant key is stored in
 *   one 128 bit key: keys[0-15]
 *
 * (NOTE: I'm not sure why, but it seems that only values 0-12 are used,
 * resulting in 104 bits of keying, not 128)
 */
void
wep_keygen128(char *str, u_char *keys)
{
    MD5_CTX ctx;
    u_char buf[64];
    int i, j;

    /* repeat str until buf is full */
    j = 0;
    for(i = 0; i < 64; i++) {
        if(str[j] == 0)
            j = 0;
        buf[i] = str[j++];
    }

    MD5_Init (&ctx);
    MD5_Update (&ctx, buf, sizeof buf);
    MD5_Final (buf, &ctx);

    memcpy(keys, buf, WEPKEYSTORE);

    for(i = 0; i < WEPSTRONGKEYSIZE; i++) {
        keys[i] = buf[i];
    }

    for(; i < WEPKEYSTORE; i++) {
        keys[i] = 0;
    }

    return;
}

/*
 * generate four subkeys from a string using the defacto standard
 *
 * resultant keys are stored in
 *   four 40 bit keys: keys[0-4], keys[5-9], keys[10-14] and keys[15-20]
 */
void
wep_keygen40(char *str, u_char *keys)
{
    int val, i, shift;

    /*
     * seed is generated by xor'ing in the keystring bytes
     * into the four bytes of the seed, starting at the little end
     */
    val = 0;
    for(i = 0; str[i]; i++) {
        shift = i & 0x3;
        val ^= (str[i] << (shift * 8));
    }

    wep_seedkeygen(val, keys);
    return;
}

void
wep_keyprint(u_char *keys)
{
    int i;
    char sepchar;

    for(i = 0; i < WEPKEYSTORE; i++) {
        sepchar = (i % WEPKEYSIZE == WEPKEYSIZE - 1) ? '\n' : ':';
        if (strong) sepchar = ':';
        printf ("%02x", keys [i]);
        if (strong && (i >= 12)) { putchar ('\n'); break; }
        putchar (sepchar);
    }
}

void wepkeygen_usage (void)
{
    printf ("wepkeygen %s", PROGNAME, REVISION);
    puts (" - Translates WEP passphrases to hex key strings");
    puts ("");
    printf ("Usage: %s    keystring", PROGNAME);
    puts (" - Generates 4 keys for  64-bit WEP");
    printf ("       %s -s keystring", PROGNAME);
    puts (" - Generates 1 key  for 128-bit WEP");
    puts ("");
    printf ("%s%s\n",
        "Colons are added to  improve readability.",
        "  These should be normally be");
    puts ("removed when the generated keys are used.");
    exit (1);
}

int main (int argc, char **argv)
{
    u_char keys [WEPKEYSTORE];
    char *genstr;
    int ch;

    strong = 0;
    while((ch = getopt(argc, argv, "s")) != EOF) {
        switch(ch) {
        case 's':
            strong = 1;
            break;
        default:
            wepkeygen_usage();
        }
    }

    argc -= optind;
    argv += optind;

    if (argc != 1) wepkeygen_usage();
    genstr = argv [0];

    if (strong)
        wep_keygen128 (genstr, keys);
    else
        wep_keygen40 (genstr, keys);

    wep_keyprint(keys);
    return 0;
}
